@@ -1,9 +1,10 @@
 # -*- coding: utf-8 -*-
-
 import datetime
-import time
 import re
 
+import time
+
+
 def formatter_st(v, default=None) -> time.struct_time:
     if v is None:
         v = default
@@ -40,22 +41,3 @@ def formatter_list(v, default=None) -> list:
         return []
     else:
         raise ValueError('the value must be str or list')
-
-
-def read_kvfile(kvfile) -> dict:
-    result = dict()
-    with open(kvfile, 'r') as f:
-        lines = f.readlines()
-        for s in lines:
-            s = s.strip()
-            if not s:
-                continue
-            p = s.find('=')
-            if p <= 0:
-                continue
-            k = s[0:p].strip()
-            v = s[p + 1:].strip()
-            if not k or not v:
-                continue
-            result[k] = v
-    return result
\ No newline at end of file
